@@ -39,19 +39,30 @@ class binner_encoding:
         #     return {'type': 'scalar', 'expression': binner.expression, 'dtype': dtype, 'count': binner.bins, 'minimum': binner.vmin, 'maximum': binner.vmax}
 
     @staticmethod
-    def decode(encoding, binner_spec):
+    def decode(encoding, binner_spec, nthreads):
         type = binner_spec['binner-type']
         dtype = encoding.decode('dtype', binner_spec['dtype'])
         if type == 'ordinal':
             cls = vaex.utils.find_type_from_dtype(vaex.superagg, "BinnerOrdinal_", dtype)
-            return cls(binner_spec['expression'], binner_spec['count'], binner_spec['minimum'])
+            return cls(nthreads, binner_spec['expression'], binner_spec['count'], binner_spec['minimum'], False)
         elif type == 'scalar':
             cls = vaex.utils.find_type_from_dtype(vaex.superagg, "BinnerScalar_", dtype)
-            return cls(binner_spec['expression'], binner_spec['minimum'], binner_spec['maximum'], binner_spec['count'])
+            return cls(nthreads, binner_spec["expression"], binner_spec["minimum"], binner_spec["maximum"], binner_spec["count"])
+        elif type == "hash":
+            cls = vaex.utils.find_type_from_dtype(vaex.superagg, "BinnerHash_", dtype)
+            hash_map_unique_id = binner_spec["hash_map_unique"]
+            if encoding.has_object(hash_map_unique_id):
+                hash_map_unique = encoding.get_object(hash_map_unique_id)
+            else:
+                hash_map_unique_spec = encoding.get_object_spec(hash_map_unique_id)
+                hash_map_unique = encoding.decode("hash-map-unique", hash_map_unique_spec)
+                encoding.set_object(hash_map_unique_id, hash_map_unique)
+            return cls(binner_spec["expression"], hash_map_unique._internal)
         else:
             raise ValueError('Cannot deserialize: %r' % binner_spec)
 
 class TaskPart:
+    stopped = False
     def __init__(self, df, expressions, name, pre_filter):
         self.df = df
         self.expressions = expressions
@@ -125,10 +136,10 @@ class TaskPartFilterFill(TaskPart):
 
 
 @register
-class TaskPartSetCreate(TaskPart):
-    snake_name = "set_create"
+class TaskPartHashmapUniqueCreate(TaskPart):
+    snake_name = "hash_map_unique_create"
 
-    def __init__(self, df, expression, dtype, dtype_item, flatten, unique_limit, selection, nthreads, return_inverse):
+    def __init__(self, df, expression, dtype, dtype_item, flatten, limit, limit_raise, selection, nthreads, return_inverse):
         expression = str(expression)
         self.nthreads = nthreads
         self.df = df
@@ -138,7 +149,8 @@ class TaskPartSetCreate(TaskPart):
         self.dtype_item = dtype_item
         self.flatten = flatten
         self.expression = expression
-        self.unique_limit = unique_limit
+        self.limit = limit
+        self.limit_raise = limit_raise
         self.selection = selection
         self.return_inverse = return_inverse
         self.chunks = []
@@ -154,12 +166,11 @@ class TaskPartSetCreate(TaskPart):
         #         transient = True
         # self.dtype = self.df.data_type(str(expression))
         # self.dtype_item = self.data_type(expression, axis=-1 if flatten else 0)
-        self.ordered_set_type = vaex.hash.ordered_set_type_from_dtype(dtype_item, transient)
         # *7 is arbitrary, but we can have more maps than threads to avoid locks
-        self.set = self.ordered_set_type(self.nthreads*7)
+        self.hash_map_unique = vaex.hash.HashMapUnique(self.dtype_item, self.nthreads*7, limit=self.limit)
 
     def get_bin_count(self):
-        return len(self.set)
+        return len(self.hash_map_unique)
 
     @property
     def expressions(self):
@@ -167,14 +178,15 @@ class TaskPartSetCreate(TaskPart):
 
     def get_result(self):
         if self.return_inverse:
-            return self.set, self.values
+            return self.hash_map_unique, self.values
         else:
-            return self.set
+            return self.hash_map_unique
 
     def process(self, thread_index, i1, i2, filter_mask, selection_masks, blocks):
         ar = blocks[0]
-        from vaex.column import _to_string_sequence
         self._check_row_limit()
+        if self.stopped:
+            return
         if self.selection:
             selection_mask = selection_masks[0]
             ar = filter(ar, selection_mask)
@@ -182,36 +194,23 @@ class TaskPartSetCreate(TaskPart):
             return
         if self.dtype.is_list and self.flatten:
             ar = ar.values
-        if self.dtype_item.is_string:
-            ar = _to_string_sequence(ar)
-        else:
-            ar = vaex.array_types.to_numpy(ar)
-            if ar.strides != (ar.itemsize,):
-                ar = ar.copy()
-        chunk_size = 1024*1024
-
+        result = self.hash_map_unique.add(ar, return_inverse=self.return_inverse)
+        if self.return_inverse:
+            values, map_index = result
+            self.chunks.append((i1, i2, values, map_index))
         self._check_row_limit()
-        if np.ma.isMaskedArray(ar):
-            mask = np.ma.getmaskarray(ar)
-            if self.return_inverse:
-                values, map_index = self.set.update(ar, mask, -1, chunk_size=chunk_size, bucket_size=chunk_size*4, return_values=self.return_inverse)
-                self.chunks.append((i1, i2, values, map_index))
-            else:
-                self.set.update(ar, mask,  -1, chunk_size=chunk_size, bucket_size=chunk_size*4)
-        else:
-            if self.return_inverse:
-                values, map_index = self.set.update(ar, -1, chunk_size=chunk_size, bucket_size=chunk_size*4, return_values=self.return_inverse)
-                self.chunks.append((i1, i2, values, map_index))
-            else:
-                self.set.update(ar, -1, chunk_size=chunk_size, bucket_size=chunk_size*4)
         if logger.level >= logging.DEBUG:
-            logger.debug(f"set uses {sys.getsizeof(self.set):,} bytes (offset {i1:,}, length {i2-i1:,})")
+            logger.debug(f"set uses {sys.getsizeof(self.hash_map_unique):,} bytes (offset {i1:,}, length {i2-i1:,})")
         self._check_row_limit()
 
     def _check_row_limit(self):
-        if self.unique_limit is not None:
-            if len(self.set) > self.unique_limit:
-                raise vaex.RowLimitException(f'Resulting set would have >= {self.unique_limit} unique combinations')
+        if self.limit is not None:
+            # we only raise when we EXCEED the limit
+            if self.limit_raise and len(self.hash_map_unique) > self.limit:
+                raise vaex.RowLimitException(f'Resulting hash_map_unique would have >= {self.limit} unique combinations')
+            # but we can stop when we are AT the limit
+            if not self.limit_raise and len(self.hash_map_unique) >= self.limit:
+                self.stopped = True
 
     def ideal_splits(self, nthreads):
         # TODO, we want to configure this
@@ -219,12 +218,12 @@ class TaskPartSetCreate(TaskPart):
 
     def reduce(self, others):
         all = [self] + others
-        all = [k.set for k in all if k.set is not None]
-        set_merged, *others = all
+        all = [k.hash_map_unique for k in all if k.hash_map_unique is not None]
+        hash_map_unique_merged, *others = all
         import time
         t0 = time.time()
-        set_merged.merge(others)
-        logger.info(f'merge took {time.time()-t0} seconds, size {len(set_merged):,}, byte_size {sys.getsizeof(set_merged):,}')
+        hash_map_unique_merged.merge(others)
+        logger.info(f'merge took {time.time()-t0} seconds, size {len(hash_map_unique_merged):,}, byte_size {sys.getsizeof(hash_map_unique_merged):,}')
 
         if self.return_inverse:
             # sort by row index
@@ -235,22 +234,27 @@ class TaskPartSetCreate(TaskPart):
             self.values = np.empty(length, vaex.dtype_of(self.chunks[0][2]).numpy)
             # TODO: we could do this parallel, but overhead is small
             for i1, i2, values, map_index in self.chunks:
-                set_merged.flatten_values(values, map_index, self.values[i1:i2])
+                hash_map_unique_merged.flatten_values(values, map_index, self.values[i1:i2])
 
-        if self.unique_limit is not None:
-            count = len(set_merged)
-            if count > self.unique_limit:
-                raise vaex.RowLimitException(f'Resulting set has {count:,} unique combinations, which is larger than the allowed value of {self.unique_limit:,}')
-        self.set = set_merged
-        self.set.fingerprint = f'set-{self.fingerprint}'
+        if self.limit is not None:
+            count = len(hash_map_unique_merged)
+            if count > self.limit:
+                if self.limit_raise:
+                    raise vaex.RowLimitException(f'Resulting set has {count:,} unique combinations, which is larger than the allowed value of {self.limit:,}')
+                else:
+                    hash_map_unique_merged = hash_map_unique_merged.limit(self.limit)
+        logger.debug("flatten hashmap...")
+        self.hash_map_unique = hash_map_unique_merged.flatten()
+        logger.debug("flatten hashmap done")
+        self.hash_map_unique._internal.fingerprint = f'hash-map-unique-{self.fingerprint}'
 
     @classmethod
     def decode(cls, encoding, spec, df, nthreads):
         return cls(df, spec['expression'], encoding.decode('dtype', spec['dtype']), encoding.decode('dtype', spec['dtype_item']),
-                   flatten=spec['flatten'], unique_limit=spec['unique_limit'], selection=spec['selection'], return_inverse=spec['return_inverse'], nthreads=nthreads)
+                   flatten=spec['flatten'], limit=spec['limit'], limit_raise=spec['limit_raise'], selection=spec['selection'], return_inverse=spec['return_inverse'], nthreads=nthreads)
 
     def memory_usage(self):
-        return sys.getsizeof(self.set)
+        return sys.getsizeof(self.hash_map_unique._internal)
 
 
 @register
@@ -478,6 +482,7 @@ class TaskPartAggregation(TaskPart):
         self.has_values = False
         self.dtypes = dtypes
         self.binners = binners
+        self.nthreads = nthreads
         # self.expressions_all = expressions
         self.expressions = [binner.expression for binner in binners]
         # TODO: selection and edges in descriptor?
@@ -491,7 +496,7 @@ class TaskPartAggregation(TaskPart):
         def create_aggregator(aggregator_descriptor, selections, initial_values):
             # for each selection, we have a separate aggregator, sharing the grid and binners
             for i, selection in enumerate(selections):
-                agg = aggregator_descriptor._create_operation(self.grid)
+                agg = aggregator_descriptor._create_operation(self.grid, nthreads)
                 self.nbytes += sys.getsizeof(agg)
                 if initial_values is not None:
                     print(np.asarray(agg))
@@ -515,18 +520,7 @@ class TaskPartAggregation(TaskPart):
         return self.nbytes
 
     def ideal_splits(self, nthreads):
-        # We need to do some proper work on this, but this should already improve performance
-        # Since if we have a lot of data per task, we should split up the work less
-        logger.info(f'A single task part takes {self.nbytes:,} bytes')
-        splits = nthreads
-        if self.nbytes >= 1e5:
-            splits = splits//2
-        if self.nbytes >= 1e6:
-            splits = splits//2
-        if self.nbytes >= 1e7:
-            splits = splits//2
-        logger.info(f'Estimate for ideal number of splits: {splits:,}')
-        return max(min(nthreads, 2), splits)
+        return 1
 
     def process(self, thread_index, i1, i2, filter_mask, selection_masks, blocks):
         # self.check()
@@ -555,19 +549,27 @@ class TaskPartAggregation(TaskPart):
         # we need to make sure we keep some objects alive, since the c++ side does not incref
         # on set_data and set_data_mask
         references = []
-        for binner in grid.binners:
+
+        def init_binner(binner):
             block = block_map[binner.expression]
             dtype = self.dtypes[binner.expression]
             block = check_array(block, dtype)
             if np.ma.isMaskedArray(block):
                 block, mask = block.data, np.ma.getmaskarray(block)
-                binner.set_data(block)
-                binner.set_data_mask(mask)
+                binner.set_data(thread_index, block)
+                binner.set_data_mask(thread_index, mask)
                 references.extend([block, mask])
             else:
-                binner.set_data(block)
-                binner.clear_data_mask()
+                binner.set_data(thread_index, block)
+                binner.clear_data_mask(thread_index, )
                 references.extend([block])
+
+        for binner in grid.binners:
+            if hasattr(binner, "binners"):
+                for subbinner in binner.binners:
+                    init_binner(subbinner)
+            else:
+                init_binner(binner)
         all_aggregators = []
 
         selection_index_global = 0
@@ -585,7 +587,10 @@ class TaskPartAggregation(TaskPart):
                     # some aggregators make a distiction between missing value and no value
                     # like nunique, they need to know if they should take the value into account or not
                     if hasattr(agg, 'set_selection_mask'):
-                        agg.set_selection_mask(selection_mask)
+                        if selection_mask is not None:
+                            agg.set_selection_mask(thread_index, selection_mask)
+                        else:
+                            agg.clear_selection_mask(thread_index)
                 selection_index_global += 1
                 if agg_desc.expressions:
                     assert len(agg_desc.expressions) in [1, 2], "only length 1 or 2 supported for now"
@@ -596,7 +601,7 @@ class TaskPartAggregation(TaskPart):
                         if np.ma.isMaskedArray(block):
                             block, mask = block.data, np.ma.getmaskarray(block)
                             block = check_array(block, dtype)
-                            agg.set_data(block, i)
+                            agg.set_data(thread_index, block, i)
                             references.extend([block])
                             if selection_mask is None:
                                 selection_mask = ~mask
@@ -605,15 +610,15 @@ class TaskPartAggregation(TaskPart):
                             references.append(selection_mask)
                         else:
                             block = check_array(block, dtype)
-                            agg.set_data(block, i)
+                            agg.set_data(thread_index, block, i)
                             references.extend([block])
                 # we only have 1 data mask, since it's locally combined
                 if selection_mask is not None:
-                    agg.set_data_mask(selection_mask)
+                    agg.set_data_mask(thread_index, selection_mask)
                     references.extend([selection_mask])
                 else:
-                    agg.clear_data_mask()
-        grid.bin(all_aggregators, N)
+                    agg.clear_data_mask(thread_index)
+        grid.bin(thread_index, all_aggregators, N)
         self.has_values = True
 
     def reduce(self, others):
@@ -621,7 +626,7 @@ class TaskPartAggregation(TaskPart):
             for selection_index, selection in enumerate(selections):
                 agg0 = aggregation[selection_index]
                 aggs = [other.aggregations[agg_index][2][selection_index] for other in others]
-                agg0.reduce(aggs)
+                agg0.merge(aggs)
 
     def get_result(self):
         results = []
@@ -654,12 +659,12 @@ class TaskPartAggregation(TaskPart):
         # aggs = [vaex.agg._from_spec(agg_spec) for agg_spec in spec['aggregations']]
         aggs = encoding.decode_list('aggregation', spec['aggregations'])
         dtypes = encoding.decode_dict('dtype', spec['dtypes'])
-        grid = encoding.decode_list('binner-cpu', spec['binners'])
+        grid = encoding.decode_list('binner-cpu', spec['binners'], nthreads=nthreads)
         values = encoding.decode_list2('ndarray', spec['values']) if 'values' in spec else None
         # dtypes = {expr: _deserialize_type(type_spec) for expr, type_spec in spec['dtypes'].items()}
         for agg in aggs:
             agg._prepare_types(df)
-        return cls(df, grid, aggs, dtypes, initial_values=values)
+        return cls(df, grid, aggs, dtypes, initial_values=values, nthreads=nthreads)
 
     def encode(self, encoding):
         # TODO: get rid of dtypes
