@@ -2,6 +2,7 @@ from __future__ import division, print_function
 import ast
 import asyncio
 from collections import defaultdict
+import contextlib
 import os
 import time
 import threading
@@ -9,6 +10,7 @@ import math
 import multiprocessing
 import logging
 import queue
+from typing import List
 
 import dask.utils
 import numpy as np
@@ -20,17 +22,17 @@ import vaex.memory
 import vaex.multithreading
 import vaex.vaexfast
 import vaex.events
+import vaex.settings
 
+try:
+    # support py 36 for the moment
+    import contextvars
+    has_contextvars = True
+except ModuleNotFoundError:
+    has_contextvars = False
 
-async_default = str(os.environ.get('VAEX_ASYNC', 'nest'))
-chunk_size_min_default = int(os.environ.get('VAEX_CHUNK_SIZE_MIN', 1024))
-chunk_size_max_default = int(os.environ.get('VAEX_CHUNK_SIZE_MAX', 1024*1024))
-chunk_size_default = ast.literal_eval(os.environ.get('VAEX_CHUNK_SIZE', 'None'))
-if chunk_size_default is not None:
-    chunk_size_default = int(chunk_size_default)  # make sure it's an int
 
 logger = logging.getLogger("vaex.execution")
-thread_count_default = multiprocessing.cpu_count()
 
 
 class UserAbort(Exception):
@@ -49,24 +51,36 @@ class Run:
             assert self.tasks[0].df._index_start == task.df._index_start
             assert self.tasks[0].df._index_end == task.df._index_end
         self.dataset = dataset
-        self.pre_filter = tasks[0].pre_filter
-        if any(self.pre_filter != task.pre_filter for task in tasks[1:]):
-            raise ValueError(f"All tasks need to be pre_filter'ed or not pre_filter'ed, it cannot be mixed: {tasks}")
-        if self.pre_filter and not all(task.df.filtered for task in tasks):
-            raise ValueError("Requested pre_filter for task while DataFrame is not filtered")
+        self.pre_filter_per_df = {}
         self.expressions = list(set(expression for task in tasks for expression in task.expressions_all))
         self.tasks_per_df = dict()
         for task in tasks:
             if task.df not in self.tasks_per_df:
                 self.tasks_per_df[task.df] = []
             self.tasks_per_df[task.df].append(task)
+        for task in tasks:
+            if task.df not in self.pre_filter_per_df:
+                self.pre_filter_per_df[task.df] = task.pre_filter
+            else:
+                if self.pre_filter_per_df[task.df] != task.pre_filter:
+                    raise ValueError(f"All tasks need to be pre_filter'ed or not pre_filter'ed, it cannot be mixed: {tasks}")
+
 
         # find the columns from the dataset we need
         dfs = set()
         for task in tasks:
             dfs.add(task.df)
-        columns = set()
+        # shared between all datasets (TODO: can we detect the filter is cached already?)
+        self.dataset_deps = set()
+        # per dataframe:
+        self.filter_deps = {}
+        self.selection_deps = {}
+        self.expression_deps = {}
+
         for df in dfs:
+            self.filter_deps[df] = set()
+            self.selection_deps[df] = set()
+            self.expression_deps[df] = set()
             others = set(df.variables) | set(df.virtual_columns) | set(df.selection_histories)
             tasks_df = [task for task in tasks if task.df == df]
             expressions = list(set(expression for task in tasks_df for expression in task.expressions_all))
@@ -74,10 +88,6 @@ class Run:
             variables = set()
             for expression in expressions:
                 variables |= df._expr(expression).expand().variables(ourself=True)
-            for selection in selections:
-                variables |= df._selection_expression(selection).dependencies()
-            if df.filtered:
-                variables |= df.get_selection(vaex.dataframe.FILTER_SELECTION_NAME).dependencies(df)
             for var in variables:
                 if var not in self.dataset:
                     if var not in others:
@@ -85,9 +95,36 @@ class Run:
                     else:
                         pass  # ok, not a column, just a var or virtual column
                 else:
-                    columns.add(var)
-        logger.debug('Using columns %r from dataset', columns)
-        self.columns = columns
+                    self.dataset_deps.add(var)
+                    self.expression_deps[df].add(var)
+
+            variables = set()
+            if df.filtered:
+                variables = df.get_selection(vaex.dataframe.FILTER_SELECTION_NAME).dependencies(df)
+                for var in variables:
+                    if var not in self.dataset:
+                        if var not in others:
+                            raise RuntimeError(f'Oops, requesting column {var} from dataset, but it does not exist')
+                        else:
+                            pass  # ok, not a column, just a var or virtual column
+                    else:
+                        self.dataset_deps.add(var)
+                        self.filter_deps[df].add(var)
+            variables = set()
+            for selection in selections:
+                if selection is not None:
+                    variables |= df._selection_expression(selection).dependencies()
+            for var in variables:
+                if var not in self.dataset:
+                    if var not in others:
+                        raise RuntimeError(f'Oops, requesting column {var} from dataset, but it does not exist')
+                    else:
+                        pass  # ok, not a column, just a var or virtual column
+                else:
+                    self.dataset_deps.add(var)
+                    self.selection_deps[df].add(var)
+
+        logger.debug('Using columns %r from dataset', self.dataset_deps)
 
 
 def _merge(tasks):
@@ -126,33 +163,53 @@ def _merge_tasks_for_df(tasks, df):
             def assign(value, i=i, subtask=subtask):
                 subtask.fulfill(value[i])
             assign(task_merged)
+            task_merged.done(None, subtask.reject)
+            task_merged.signal_start.connect(subtask.signal_start.emit)
         tasks_merged.append(task_merged)
     return tasks_non_mergable + tasks_merged
 
 
 class Executor:
     """An executor is responsible to executing tasks, they are not reentrant, but thread safe"""
-    def __init__(self, async_method=async_default):
-        self.tasks = []
+    def __init__(self, async_method=None):
+        self.tasks : List[Task] = []
         self.async_method = async_method
         self.signal_begin = vaex.events.Signal("begin")
         self.signal_progress = vaex.events.Signal("progress")
         self.signal_end = vaex.events.Signal("end")
         self.signal_cancel = vaex.events.Signal("cancel")
         self.local = threading.local()  # to make it non-reentrant
+        if has_contextvars:
+            # used for calling execute_async from different async callstacks
+            self.isnested = contextvars.ContextVar('executor', default=False)
         self.lock = threading.Lock()
         self.event_loop = asyncio.new_event_loop()
 
+    if hasattr(contextlib, 'asynccontextmanager'):
+        @contextlib.asynccontextmanager
+        async def auto_execute(self):
+            '''This async executor will start executing tasks automatically when a task is awaited for.'''
+            vaex.promise.auto_await_executor.set(self)
+            try:
+                yield
+                await self.execute_async()
+            finally:
+                vaex.promise.auto_await_executor.set(None)
+    else:
+        async def auto_execute(self):
+            raise NotImplemented('Only on Python >3.7')
+
     async def execute_async(self):
-        pass
+        raise NotImplementedError
 
     def execute(self):
-        return self.run(self.execute_async())
+        raise NotImplementedError
 
     def run(self, coro):
-        if self.async_method == "nest":
+        async_method = self.async_method or vaex.settings.main.async_
+        if async_method == "nest":
             return vaex.asyncio.just_run(coro)
-        elif self.async_method == "awaitio":
+        elif async_method == "awaitio":
             with vaex.asyncio.with_event_loop(self.event_loop):
                 return self.event_loop.run_until_complete(coro)
         else:
@@ -206,12 +263,12 @@ class Executor:
 
 
 class ExecutorLocal(Executor):
-    def __init__(self, thread_pool=None, chunk_size=chunk_size_default, chunk_size_min=chunk_size_min_default, chunk_size_max=chunk_size_max_default, thread_mover=None, zigzag=True):
+    def __init__(self, thread_pool=None, chunk_size=None, chunk_size_min=None, chunk_size_max=None, thread_mover=None, zigzag=True):
         super().__init__()
         self.thread_pool = thread_pool or vaex.multithreading.ThreadPoolIndex()
         self.chunk_size = chunk_size
-        self.chunk_size_min = chunk_size_min_default
-        self.chunk_size_max = chunk_size_max_default
+        self.chunk_size_min = chunk_size_min
+        self.chunk_size_max = chunk_size_max
         self.thread = None
         self.passes = 0  # how many times we passed over the data
         self.zig = True  # zig or zag
@@ -224,15 +281,34 @@ class ExecutorLocal(Executor):
         run.cancelled = True
 
     def chunk_size_for(self, row_count):
-        chunk_size = self.chunk_size
+        chunk_size = self.chunk_size or vaex.settings.main.chunk.size
+        chunk_size_min = self.chunk_size or vaex.settings.main.chunk.size_min
+        chunk_size_max = self.chunk_size or vaex.settings.main.chunk.size_max
         if chunk_size is None:
             # we determine it automatically by defaulting to having each thread do 1 chunk
             chunk_size_1_pass = vaex.utils.div_ceil(row_count, self.thread_pool.nthreads)
             # brackated by a min and max chunk_size
-            chunk_size = min(self.chunk_size_max, max(self.chunk_size_min, chunk_size_1_pass))
+            chunk_size = min(chunk_size_max, max(chunk_size_min, chunk_size_1_pass))
         return chunk_size
 
     async def execute_async(self):
+        # consume awaitables from the generator, and await them here at the top
+        # level, such that we don't need await in the downstream code
+        # so we can reuse the same code in a sync way
+        gen = self.execute_generator(use_async=True)
+        value = None
+        while True:
+            try:
+                value = gen.send(value)
+                value = await value
+            except StopIteration:
+                break
+
+    def execute(self):
+        for _ in self.execute_generator():
+            pass  # just eat all elements
+
+    def execute_generator(self, use_async=False):
         logger.debug("starting with execute")
 
         with self.lock:  # setup thread local initial values
@@ -243,11 +319,10 @@ class ExecutorLocal(Executor):
             t0 = time.time()
             self.local.cancelled = False
             self.signal_begin.emit()
-            cancelled = False
             # keep getting a list of tasks
             # we currently process tasks (grouped) per df
             # but also, tasks can add new tasks
-            while not cancelled:
+            while True:
                 tasks = self.local.tasks = self._pop_tasks()
 
                 # wo don't allow any thread from our thread pool to enter (a computation should never produce a new task)
@@ -255,13 +330,15 @@ class ExecutorLocal(Executor):
                 chunk_executor_thread = threading.current_thread() in self.thread_pool._threads
                 import traceback
                 trace = ''.join(traceback.format_stack())
-                if chunk_executor_thread or self.local.executing:
+                if chunk_executor_thread or self.local.executing and (has_contextvars is False or self.isnested.get() is True):
                     logger.error("nested execute call")
                     raise RuntimeError("nested execute call: %r %r\nlast trace:\n%s\ncurrent trace:\n%s" % (chunk_executor_thread, self.local.executing, self.local.last_trace, trace))
                 else:
                     self.local.last_trace = trace
 
                 self.local.executing = True
+                if has_contextvars:
+                    self.isnested.set(True)
 
                 if not tasks:
                     break
@@ -285,6 +362,9 @@ class ExecutorLocal(Executor):
                             task.reject(e)
                             raise
 
+                for task in run.tasks:
+                    task.signal_start.emit(self)
+
                 for task in run.tasks:
                     task._results = []
                     if not any(task.signal_progress.emit(0)):
@@ -338,34 +418,30 @@ class ExecutorLocal(Executor):
                 # if self.zigzag:
                 #     self.zig = not self.zig
                 def progress(p):
-                    return all(self.signal_progress.emit(p)) and\
-                           all([all(task.signal_progress.emit(p)) for task in tasks]) and\
-                           all([not task.cancelled for task in tasks])
-                async for _element in self.thread_pool.map_async(self.process_part, dataset.chunk_iterator(run.columns, chunk_size),
+                    # no global cancel and at least 1 tasks wants to continue, then we continue
+                    ok_tasks = any([task.progress(p) for task in tasks])
+                    all_stopped = all([task.stopped for task in tasks])
+                    ok_executor = all(self.signal_progress.emit(p))
+                    if all_stopped:
+                        logger.debug("Pass cancelled because all tasks are stopped: %r", tasks)
+                    if not ok_tasks:
+                        logger.debug("Pass cancelled because all tasks cancelled: %r", tasks)
+                    if not ok_executor:
+                        logger.debug("Pass cancelled because of the global progress event: %r", self.signal_progress.callbacks)
+                    return ok_tasks and ok_executor and not all_stopped
+                yield from self.thread_pool.map(self.process_part, dataset.chunk_iterator(run.dataset_deps, chunk_size),
                                                     dataset.row_count,
                                                     progress=progress,
-                                                    cancel=lambda: self._cancel(run), unpack=True, run=run):
-                    pass  # just eat all element
+                                                    cancel=lambda: self._cancel(run), unpack=True, run=run, use_async=use_async)
                 duration_wallclock = time.time() - t0
                 logger.debug("executing took %r seconds", duration_wallclock)
-                cancelled = self.local.cancelled or any(task.cancelled for task in tasks) or run.cancelled
-                logger.debug("cancelled: %r", cancelled)
                 self.local.executing = False
-                if cancelled:
-                    logger.debug("execution aborted")
+                if has_contextvars:
+                    self.isnested.set(False)
+                if True:  # kept to keep the diff small
                     for task in tasks:
-                        task.reject(UserAbort("cancelled"))
-                        # remove references
-                        task._result = None
-                        task._results = None
-                        cancelled = True
-                        if isinstance(task, vaex.tasks.TaskAggregations):
-                            for subtask in task.original_tasks:
-                                subtask.reject(UserAbort("cancelled"))
-                else:
-                    for task in tasks:
-                        logger.debug("fulfill task: %r", task)
                         if not task.cancelled:
+                            logger.debug("fulfill task: %r", task)
                             parts = task._parts
                             if not isinstance(parts, list):
                                 parts_queue = parts
@@ -398,9 +474,13 @@ class ExecutorLocal(Executor):
                                         logger.info("added result: %r in cache under key: %r", task_cachable.get(), key)
 
                         else:
-                            task.reject(UserAbort("Task was cancelled"))
+                            logger.debug("rejecting task: %r", task)
+                            # we now reject, in the main thread
+                            if task._toreject:
+                                task.reject(task._toreject)
+                            else:
+                                task.reject(UserAbort("Task was cancelled"))
                             # remove references
-                            cancelled = True
                         task._result = None
                         task._results = None
                     self.signal_end.emit()
@@ -409,6 +489,8 @@ class ExecutorLocal(Executor):
             raise
         finally:
             self.local.executing = False
+            if has_contextvars:
+                self.isnested.set(False)
 
     def process_part(self, thread_index, i1, i2, chunks, run):
         if not run.cancelled:
@@ -429,13 +511,15 @@ class ExecutorLocal(Executor):
             assert df in run.variables
             from .scopes import _BlockScope
 
-            if run.pre_filter:
-                # TODO: move detection of dependencies up
-                filter_deps = df.get_selection(vaex.dataframe.FILTER_SELECTION_NAME).dependencies(df)
+            expressions = list(set(expression for task in tasks for expression in task.expressions_all))
+            pre_filter = tasks[0].pre_filter
+            if pre_filter:
+                filter_deps = run.filter_deps[df]
                 filter_scope = _BlockScope(df, i1, i2, None, selection=True, values={**run.variables[df], **{k: chunks[k] for k in filter_deps if k in chunks}})
                 filter_scope.filter_mask = None
                 filter_mask = filter_scope.evaluate(vaex.dataframe.FILTER_SELECTION_NAME)
-                chunks = {k:vaex.array_types.filter(v, filter_mask) for k, v, in chunks.items()}
+                deps = run.expression_deps[df] | run.selection_deps[df]
+                chunks = {k:vaex.array_types.filter(v, filter_mask) for k, v, in chunks.items() if k in deps}
             else:
                 filter_mask = None
             def sanity_check(name, chunk):
@@ -445,16 +529,17 @@ class ExecutorLocal(Executor):
                 sanity_check(name, chunk)
             block_scope = _BlockScope(df, i1, i2, values={**run.variables[df], **chunks})
             block_scope.mask = filter_mask
-            expressions = list(set(expression for task in tasks for expression in task.expressions_all))
             block_dict = {expression: block_scope.evaluate(expression) for expression in expressions}
             selection_scope = _BlockScope(df, i1, i2, None, selection=True, values={**block_scope.values})
             selection_scope.filter_mask = filter_mask
-            if not run.pre_filter and df.filtered:
+            if not pre_filter and df.filtered:
                 filter_mask = selection_scope.evaluate(vaex.dataframe.FILTER_SELECTION_NAME)
 
             memory_tracker = vaex.memory.create_tracker()
             task_checkers = vaex.tasks.create_checkers()
             for task in tasks:
+                if task.stopped:
+                    continue
                 assert df is task.df
                 blocks = [block_dict[expression] for expression in task.expressions_all]
                 def fix(ar):
@@ -471,19 +556,30 @@ class ExecutorLocal(Executor):
                             task_parts = [task._parts[thread_index]]
                         else:
                             task_parts = [task._parts.get()]
+                    some_parts_stopped = False
                     for task_index, task_part in enumerate(task_parts):
-                        try:
-                            if task.see_all:
-                                task_part.process(task_index, i1, i2, filter_mask, selections, blocks)
-                            else:
-                                task_part.process(thread_index, i1, i2, filter_mask, selections, blocks)
-                        except Exception as e:
-                            task.reject(e)
-                            run.cancelled = True
-                            raise
-                        finally:
-                            if not isinstance(task._parts, list):
-                                task._parts.put(task_part)
+                        if not task_part.stopped:
+                            try:
+                                if task.see_all:
+                                    task_part.process(thread_index, i1, i2, filter_mask, selections, blocks)
+                                else:
+                                    task_part.process(thread_index, i1, i2, filter_mask, selections, blocks)
+                            except Exception as e:
+                                # we cannot call .reject, since then we'll handle fallbacks in this thread
+                                task._toreject = e
+                                task.cancelled = True
+                            finally:
+                                if not isinstance(task._parts, list):
+                                    task._parts.put(task_part)
+                            # we could be done after processing
+                            if task_part.stopped:
+                                some_parts_stopped = True
+                        else:
+                            some_parts_stopped = True
+                        if some_parts_stopped:
+                            break
+                    if some_parts_stopped:  # if 1 is done, the whole task is done
+                        task.stopped = True
                 if memory_tracker.track_live:
                     for part in task_parts:
                         memory_tracker.using(part.memory_usage())
