@@ -81,6 +81,7 @@ class Task(vaex.promise.Promise):
         self.progress_fraction = 0
         self.signal_progress.connect(self._set_progress)
         self.cancelled = False
+        self.stopped = False  # a task can stop early, but without error
         self.name = name
         self.pre_filter = pre_filter
         self.result = None
@@ -172,16 +173,17 @@ class TaskFilterFill(Task):
         return cls(df)
 
 @register
-class TaskSetCreate(Task):
+class TaskHashmapUniqueCreate(Task):
     see_all = True
     requires_fingerprint = True
-    snake_name = "set_create"
-    def __init__(self, df, expression, flatten, unique_limit=None, selection=None, return_inverse=False):
+    snake_name = "hash_map_unique_create"
+    def __init__(self, df, expression, flatten, limit=None, limit_raise=True, selection=None, return_inverse=False):
         super().__init__(df=df, expressions=[expression], pre_filter=df.filtered, name=self.snake_name)
         self.flatten = flatten
         self.dtype = self.df.data_type(expression)
         self.dtype_item = self.df.data_type(expression, axis=-1 if flatten else 0)
-        self.unique_limit = unique_limit
+        self.limit = limit
+        self.limit_raise = limit_raise
         self.selection = selection
         self.selections = [self.selection]
         self.return_inverse = return_inverse
@@ -194,7 +196,8 @@ class TaskSetCreate(Task):
 
     def encode(self, encoding):
         return {'expression': self.expressions[0], 'dtype': encoding.encode('dtype', self.dtype),
-                'dtype_item': encoding.encode('dtype', self.dtype_item), 'flatten': self.flatten, 'unique_limit': self.unique_limit,
+                'dtype_item': encoding.encode('dtype', self.dtype_item), 'flatten': self.flatten,
+                'limit': self.limit, 'limit_raise': self.limit_raise,
                 'selection': self.selection, 'return_inverse': self.return_inverse}
 
     @classmethod
@@ -423,6 +426,7 @@ class TaskStatistic(Task):
 class TaskAggregations(Task):
     """Multiple aggregations on a single grid."""
     snake_name = "aggregations"
+    see_all = True
 
     def __init__(self, df, binners):
         expressions = [binner.expression for binner in binners]
