@@ -17,8 +17,10 @@ from typing import MutableMapping
 import warnings
 import numbers
 import keyword
+from filelock import FileLock
 import pkg_resources
 
+import blake3
 import dask.utils
 import numpy as np
 import pyarrow as pa
@@ -816,7 +818,23 @@ def required_dtype_for_max(N, signed=True):
         if N <= np.iinfo(dtype).max:
             return np.dtype(dtype)
     else:
-        raise ValueError(f'Cannot store a max value on {N} inside an uint64/int64')
+        raise ValueError(f"Cannot store a max value of {N} inside an uint64/int64")
+
+
+def required_dtype_for_int(value, signed=True):
+    if signed or value < 0:
+        dtypes = [np.int8, np.int16, np.int32, np.int64]
+    else:
+        dtypes = [np.uint8, np.uint16, np.uint32, np.uint64]
+    for dtype in dtypes:
+        if value >= 0:
+            if value <= np.iinfo(dtype).max:
+                return np.dtype(dtype)
+        else:
+            if value >= np.iinfo(dtype).min:
+                return np.dtype(dtype)
+    else:
+        raise ValueError(f"Cannot store a value of {value} inside an uint64/int64")
 
 
 def print_stack_trace(*args, **kwargs):
@@ -941,7 +959,41 @@ def dropnan(sequence, expect=None):
     return original_type(non_nan)
 
 
+def dict_replace_key(d, key_old, key_new):
+    '''Replace a key, without changing order'''
+    d_new = {}
+    for key, value in d.items():
+        if key == key_old:
+            key = key_new
+        d_new[key] = value
+    return d_new
+
 # backwards compatibility
 def progressbars(*args, **kwargs):
     from .progress import tree
     return tree(*args, **kwargs)
+
+
+import contextlib
+@contextlib.contextmanager
+def file_lock(name):
+    '''Context manager for creating a file lock in the file lock directory.
+
+    :param name: A unique name for the context (e.g. a fingerprint) on which the filename is based.
+    '''
+    path = os.path.join(vaex.settings.main.path_lock, name)
+    os.makedirs(os.path.dirname(path), exist_ok=True)
+    with FileLock(path):
+        yield
+
+
+def create_hasher(initial_data=None, large_data=True):
+    """Creates a blake3 hasher"""
+    if not hasattr(blake3, "__version__"):
+        version = (0, 2)
+    else:
+        version = tuple(map(int, blake3.__version__.split(".")[:2]))
+    if version < (0, 3):
+        return blake3.blake3(initial_data, multithreading=large_data)
+    else:
+        return blake3.blake3(initial_data, max_threads=blake3.blake3.AUTO if large_data else 1)
